C The file holds a number of simple tools
C that we use in our main programs
C
C     - ElapsedTime
C	- SeqaD
C     - GetD
C     - GetI
C     - GetC
C     - DMachEps
C     - MyDate
C     - MarkovAR
C****************************************************************
C
C ElapsedTime
C
C Purpose: returns an elapsed time string
C
C Usage: Call ElapsedTime(Seconds,Time)
C
C Input:   Seconds: Real(4), gives the
C          elapsed time in seconds
C
C
C Output:  Time: Character(100), holds the
C          formated time string
C          
      Subroutine ElapsedTime(Seconds,Time)

      Real Seconds, totdays, tothrs, totmins, secs
	Integer days, hrs, mins
	Character(256) Time
	
	totdays = Seconds/(3600*24)
      days = floor(totdays)
	
      tothrs = (totdays-Real(days))*24
      hrs = floor(tothrs)
	
      totmins = (tothrs-real(hrs))*60
      mins = floor(totmins)
	
      secs = (totmins-real(mins))*60;
      
      if (days .gt. 1) then
        Write(Time,"(I1,' Days')") days
      endif

	if (days .eq. 1) then    
        Write(Time,"(I1,' Day')") days
      endif

      if (days .eq. 0) then
        Time=""
      endif    
      
	if (hrs .eq. 1) then
	  write(Time,102) trim(Time), hrs, ' hour'
      endif
102   format(A,I2,A)        
	if (hrs .gt. 1) then
	  write(Time,102) trim(Time), hrs, ' hours'
      endif
	
	if (mins .eq. 1) then
	  write(Time,102) trim(Time), mins, ' minute'
      endif

      if (mins .gt. 1) then
	  write(Time,102) trim(Time), mins, ' minutes'
      endif
      
	write(Time,103) trim(Time), secs, ' seconds'
103	format(A,F6.2,A)
            

      RETURN
	END Subroutine

C*********************************************************************
C SeqaD
C
C Puropse: creates an additive series of double precion
C
C Usage Call SeqaD(xmin, step, nobs,x)
C
C Inputs:
C 
C xmin: Real(8), the first element in the series
C step: Real(8), increment
C nobs: Integer(4), number of elements in the series
C
C Output:
C
C x: Real(8) vector with nobs elements, where
C   x(i)=xmin+(i-1)*step


      Subroutine SeqaD(xmin,step,nobs,x)

	Integer(4) nobs
      Real(8)    xmin, step, x(nobs)
	      
	x(1)=xmin
	
	do i=2,nobs
	 x(i)=x(i-1)+step	
      end do     
      
	Return
	End Subroutine

C******************************************************************
C
C GetD
C
C Purpose: read a double precion variable from a text file
C
C Usage: Call GetD(VarName,FileName,Var)
C
C GetD(VarName,FileName,Var)
C
C Input: VarName:  Character*(*), the name of the variable
C                  in the file
C        FileName: Character*(*), the name of the file
C                  (must be in the current working directory
C
C Output: Var: the name of the variable that is asigned the
C              number read from the file
C
C Example: Suppose in the ASCII-File Parameters.txt
C          contains the following lines\
C          alpha=0.35
C          beta=0.99
C          delta=0.01
C          The command CALL GetD("delta","Parameters.txt",delta)
C          returns 0.01 and stores that value in the variable
C          delta.
C

      Subroutine GetD(String1,String2,Pvalue)
 
      use DFLIB

      Character*(*), INTENT(IN) :: String1, String2
	
      Real(8) Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2, I3
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=10,file=String2,err=104)
      
      
      Do while ((.not. eof(10)) .and. (I1 .eq. 0))
	    Read(10,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in file " 
     +        //  String2 // "!"
	I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
      
      endif
      close(10)
104   If (I1 .eq. 0) then
        Text="File not found"
	  I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
	End If
      Return      
	End Subroutine
	
C**********************************************************************
C
C GetI
C
C Purpose: Same as GetD, but returns an Integer value in
C Var
C
C Usage: Call GetI(VarName,FileName,Var)

      Subroutine GetI(String1,String2,Pvalue)
 
      use DFLIB

      Character*(*), INTENT(IN) :: String1, String2
	
      Integer Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2, I3
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=10,action="Read",file=String2,err=105)
          
      Do while ((.not. eof(10)) .and. (I1 .eq. 0))
	    Read(10,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in File " 
     +        //  String2 // " !"
      I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
      endif
      close(10)
105   If (I1 .eq. 0) then
      Text="File not found"
	I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
	End If
      Return
	End Subroutine

C********************************************************************
C GetC
C
C Purpose: read a Character variable from a text file
C
C Usage: Call GetC(VarName,FileName,Var)
C
C
C Input: VarName:  Character*(*), the name of the variable
C                  in the file
C        FileName: Character*(*), the name of the file
C                  (must be in the current working directory
C
C Output: Var: the name of the variable that is asigned the
C              string read from the file
C
C Example: Suppose in the ASCII-File Parameters.txt
C          contains the following lines\
C          alpha=0.35
C          beta=0.99
C          file=outfile
C          The command CALL GetC("file","Parameters.txt",file)
C          returns the string 'outfile' in the variable file
C

      Subroutine GetC(String1,String2,String3)
 
      use DFLIB

      Character*(*), INTENT(IN) :: String1, String2
	Character*(*), INTENT(OUT) :: String3
      
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=10,file=String2,err=104)
      
      
      Do while ((.not. eof(10)) .and. (I1 .eq. 0))
	    Read(10,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) String3
      End If
      end do
      if (I1 .eq. 0) then
	continue
	String3=""
      Text="Variable " // String1 // " not found in file " 
     +        //  String2 // "!"
	I3=MESSAGEBOXQQ(Text,"Error",MB$OK)      
      endif
      close(10)
104   If (I1 .eq. 0) then
        Text="File not found"
	  I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
	End If
      Return      
	End Subroutine
	

C********************************************************************
C
C DMachEps
C
C Purpose: Returns the machine epsilon (Real(8))
C
C Usage: X=MachEps;

      Real(8) Function DMachEps

	Real(8) m
	m=1.0
	do while ((1.0+m) .gt. 1.0)
	  m=0.5*m
	end do
	DMachEps=2.0*m

	END FUNCTION
C**********************************************************************
C
C MyDate
C
C Purpose: Retrieve the current date and time and
C          returns it in a Character(50) string
C
C Usage: X=MyDate

      CHARACTER(50) Function  MyDate()

	INTEGER(2) year,month,day, hour, min, sec, hsec
	CALL GETDAT(year,month,day)
	CALL GETTIM (hour, min, sec, hsec) 
	WRITE(MyDate,"(I2,'.',I2,'.',I4,'  ',I2,' Uhr ',I2,' Minuten und ',
     +      I2,' Sekunden')") day,month,year,hour,min,sec

      Return
	END FUNCTION

C**********************************************************************
C
C MarkovAR
C
C Purpose: Approximate AR-Process by Markov Chain
C
C Usage: Call MarkovAR(size,m,rho,sigma,z,p)
C
C Input: size, scalar, Real(8), fraction of
C                      unconditional variance of AR process,
C                      determines grid size
C
C        m, Integer(4), number of states of Markov Chain
C
C      rho, Real(8), autoregressiv parameter
C  
C    sigma, Real(8), standard deviation of innovations
C
C Output: z, m x 1, Real(8) vector, states of the Markov Chain
C
C         p, m x m, Real(8) transition matrix
C
      Subroutine MarkovAR(size,m,rho,sigma,z,p)

      use IMSLF90       

	Integer(4) m
	Real(8) size, rho, sigma, z(1:m), p(1:m,1:m)

      Real(8) sigmaz, zbar
	Integer(4) I
		 
      sigmaz=dsqrt(sigma**2/(1-(rho**2)))
      zbar=size*sigmaz

      Call SeqaD(-zbar,dble(2.0*zbar/(m-1)), m,z)
  
      Do I=1,m
	
         p(i,1)=dnordf((z(1)-rho*z(i)+(z(2)-z(1))/2)/sigma)
C         p(i,m)=1-dnordf((z(m)-rho*z(i)-(z(m)-z(m-1))/2)/sigma)
         
	   Do j=2,m-1
   
           p(i,j)=dnordf((z(j)-rho*z(i)+(z(j)-z(j-1))/2)/sigma)-
     +             dnordf((z(j)-rho*z(i)-(z(j)-z(j-1))/2)/sigma)
         
	   End Do
	   p(i,m)=1-SUM(p(i,1:m-1))
	   
	End Do  
      End Subroutine
