C The file holds a number of simple tools
C that we use in our main programs
C
C     - ElapsedTime
C	- SeqaD
C     - GetD
C     - GetI
C     - GetC
C     - DMachEps
C     - MyDate
C     - HpFilter
C     - FixvMN
C     - FixpMN
C     - Corrx
C     - MarkovAR
C     - MarkovTP
C****************************************************************
C
C ElapsedTime
C
C Purpose: returns an elapsed time string
C
C Usage: Call ElapsedTime(Seconds,Time)
C
C Input:   Seconds: Real(4), gives the
C          elapsed time in seconds
C
C
C Output:  Time: Character(100), holds the
C          formated time string
C          
      Subroutine ElapsedTime(Seconds,Time)

      Real Seconds, totdays, tothrs, totmins, secs
	Integer days, hrs, mins
	Character(256) Time
	
	totdays = Seconds/(3600*24)
      days = floor(totdays)
	
      tothrs = (totdays-Real(days))*24
      hrs = floor(tothrs)
	
      totmins = (tothrs-real(hrs))*60
      mins = floor(totmins)
	
      secs = (totmins-real(mins))*60;
      
      if (days .gt. 1) then
        Write(Time,"(I1,' Days')") days
      endif

	if (days .eq. 1) then    
        Write(Time,"(I1,' Day')") days
      endif

      if (days .eq. 0) then
        Time=""
      endif    
      
	if (hrs .eq. 1) then
	  write(Time,102) trim(Time), hrs, ' hour'
      endif
102   format(A,I2,A)        
	if (hrs .gt. 1) then
	  write(Time,102) trim(Time), hrs, ' hours'
      endif
	
	if (mins .eq. 1) then
	  write(Time,102) trim(Time), mins, ' minute'
      endif

      if (mins .gt. 1) then
	  write(Time,102) trim(Time), mins, ' minutes'
      endif
      
	write(Time,103) trim(Time), secs, ' seconds'
103	format(A,F6.2,A)
            

      RETURN
	END Subroutine

C*********************************************************************
C SeqaD
C
C Puropse: creates an additive series of double precion
C
C Usage Call SeqaD(xmin, step, nobs,x)
C
C Inputs:
C 
C xmin: Real(8), the first element in the series
C step: Real(8), increment
C nobs: Integer(4), number of elements in the series
C
C Output:
C
C x: Real(8) vector with nobs elements, where
C   x(i)=xmin+(i-1)*step


      Subroutine SeqaD(xmin,step,nobs,x)

	Integer(4) nobs
      Real(8)    xmin, step, x(nobs)
	      
	x(1)=xmin
	
	do i=2,nobs
	 x(i)=x(i-1)+step	
      end do     
      
	Return
	End Subroutine

C******************************************************************
C
C GetD
C
C Purpose: read a double precion variable from a text file
C
C Usage: Call GetD(VarName,FileName,Var)
C
C GetD(VarName,FileName,Var)
C
C Input: VarName:  Character*(*), the name of the variable
C                  in the file
C        FileName: Character*(*), the name of the file
C                  (must be in the current working directory
C
C Output: Var: the name of the variable that is asigned the
C              number read from the file
C
C Example: Suppose in the ASCII-File Parameters.txt
C          contains the following lines\
C          alpha=0.35
C          beta=0.99
C          delta=0.01
C          The command CALL GetD("delta","Parameters.txt",delta)
C          returns 0.01 and stores that value in the variable
C          delta.
C

      Subroutine GetD(String1,String2,Pvalue)
      
	use DFLIB

      Character*(*), INTENT(IN) :: String1, String2
	
      Real(8) Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2, I3
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=10,file=String2,err=104)
      
      
      Do while ((.not. eof(10)) .and. (I1 .eq. 0))
	    Read(10,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in file " 
     +        //  String2 // "!"
	I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
      
      endif
      close(10)
104   If (I1 .eq. 0) then
        Text="File not found"
	  I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
	End If
      Return      
	End Subroutine
	
C**********************************************************************
C
C GetI
C
C Purpose: Same as GetD, but returns an Integer value in
C Var
C
C Usage: Call GetI(VarName,FileName,Var)

      Subroutine GetI(String1,String2,Pvalue)
 
      use DFLIB

      Character*(*), INTENT(IN) :: String1, String2
	
      Integer Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2, I3
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=10,action="Read",file=String2,err=105)
          
      Do while ((.not. eof(10)) .and. (I1 .eq. 0))
	    Read(10,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in File " 
     +        //  String2 // " !"
      I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
      endif
      close(10)
105   If (I1 .eq. 0) then
      Text="File not found"
	I3=MESSAGEBOXQQ(Text,"Error",MB$OK)
	End If
      Return
	End Subroutine

C********************************************************************
C GetC
C
C Purpose: read a Character variable from a text file
C
C Usage: Call GetC(VarName,FileName,Var)
C
C
C Input: VarName:  Character*(*), the name of the variable
C                  in the file
C        FileName: Character*(*), the name of the file
C                  (must be in the current working directory
C
C Output: Var: the name of the variable that is asigned the
C              string read from the file
C
C Example: Suppose in the ASCII-File Parameters.txt
C          contains the following lines\
C          alpha=0.35
C          beta=0.99
C          file=outfile
C          The command CALL GetC("file","Parameters.txt",file)
C          returns the string 'outfile' in the variable file
C

      Subroutine GetC(String1,String2,String3)
 
      use DFLIB
	
      Character*(*), INTENT(IN) :: String1, String2
	Character*(*), INTENT(OUT) :: String3
      
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=10,file=String2,err=104)
      
      
      Do while ((.not. eof(10)) .and. (I1 .eq. 0))
	    Read(10,'(A)') Text	    
	    Rest=Text(I2+1:len(Text))
	    Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) String3
      End If
      end do
      if (I1 .eq. 0) then
	continue
	String3=""
      Text="Variable " // String1 // " not found in file " 
     +        //  String2 // "!"
	I2=MESSAGEBOXQQ(Text,"Error",MB$OK)
      
      endif
      close(10)
104   If (I1 .eq. 0) then
        Text="File not found"
	  I2=MESSAGEBOXQQ(Text,"Error",MB$OK)
	End If
      Return      
	End Subroutine
	


C********************************************************************
C
C DMachEps
C
C Purpose: Returns the machine epsilon (Real(8))
C
C Usage: X=MachEps;

      Real(8) Function DMachEps

	Real(8) m
	m=1.0
	do while ((1.0+m) .gt. 1.0)
	  m=0.5*m
	end do
	DMachEps=2.0*m

	END FUNCTION
C**********************************************************************
C
C MyDate
C
C Purpose: Retrieve the current date and time and
C          returns it in a Character(50) string
C
C Usage: X=MyDate

      CHARACTER(50) Function  MyDate()

	INTEGER(2) year,month,day, hour, min, sec, hsec
	CALL GETDAT(year,month,day)
	CALL GETTIM (hour, min, sec, hsec) 
	WRITE(MyDate,"(I2,'.',I2,'.',I4,'  ',I2,' Uhr ',I2,' Minuten und ',
     +      I2,' Sekunden')") day,month,year,hour,min,sec

      Return
	END FUNCTION

C***********************************************************************
C
C HPFilter
C
C Purpose: Compute the Hodrick-Prescott cyclical filter
C
C Usage:   CALL HPFilter(N,Y,Lambda,C)
C
C Input: N: Integer(4), the number of observations in Y
C        Y: Real(8), column vector of length N, the series whose
C                    cyclical component is to be computed
C   Lambda: Smoothing Parameter
C
C Ouput: C: Real(8), the cyclical component if Y

      Subroutine HPFilter(N,Y,Lambda,C)

      use IMSLF90 
      
	Integer(4) N
	Real(8) Y(N), Lambda, C(N)

	Integer I, NCODA, LDA
	Real(8) M(3,N)

      Parameter (LDA=3, NCODA=2)

C     Set up band matrix
      Do I=1,N
	   M(1,I)=Lambda
	   M(2,I)=-4*Lambda
	   M(3,I)=1+6*Lambda
	End Do

	M(1,1)=0.0
	M(1,2)=0.0
	M(2,1)=0.0
	M(2,2)=-2*Lambda
	M(2,N)=-2*Lambda
	M(3,1)=1+Lambda
	M(3,2)=1+5*Lambda
	M(3,N)=1+Lambda
	M(3,N-1)=1+5*Lambda
	
	CALL DLSAQS(N,M,LDA,NCODA,Y,C)
	C=Y-C

	End Subroutine

C*********************************************************************
C
C FixvMN
C 
C Purpose: Find the solution to f(x)=0
C
C Usage: Call FixvMN(N,x1,bounds,F,x2)
C
C Input:  N, Integer(4), number of elements in x
C        x1, Real(8) x1(N) column vector
C    bounds, Real(8), N x 2 matrix with lower and upper bounds on x
C         F, pointer the function that computes f(x),
C         the call must be CALL F(N,x,fx), where fx is N x 1
C
C Output: x2, Real(8), N x 1 vector with solution.

      Subroutine FixvMN(N,x0,bounds,F,x2)

      use IMSLF90 
      Integer(4) N
	Real(8) x0(N), x1(N), bounds(1:N,1:2), x2(N)

	Integer MaxIt, itn
	Real(8) stopc, crit, xscale(N), epsfcn, df(N,N), fx(N), dx(N),
     +        step, CheckBounds2, s
	Logical CheckBounds1

	External F

C	Initialize 
      maxit=5000     ! stop after 5000 Iterations 
      stopc=1.0e-8   ! stopping criterium 
      xscale=1.0     ! no scaling
	epsfcn=0.0
	x1=x0
      itn=1           
C Start Iterations
      crit=1.0
      do while ((crit .gt. stopc) .and. (itn .lt. maxit))
	   CALL F(N,N,x1,fx)
         Call DFDJAC(F,N,N,x1,xscale,fx,epsfcn,df,N)		  
         Call DLSARG(N,df,N,-fx,1,dx)      
	   x2=x1+dx
         step=1
	   s=CheckBounds2(N,x1,dx,bounds)
         do while (CheckBounds1(N,x2,bounds))
          step=0.5*step         
          x2=x1+step*dx
         End Do
         CALL F(N,N,x2,fx)
	   fx=abs(fx)
	   crit=fx(IDMAX(N,fx,1))
         x1=x2;
        itn=itn+1
      End Do
      if (itn .gt. maxit) then
	   print *, 'Maximum number of iterations exceeded'
	   print *, 'Returned solution may be missleaden'
      endif

      End Subroutine

C*********************************************************************
C
C CheckBounds1:
C
C Purpose: Checks whether the next step computed from FixVMN exceeds
C          the allowed bounds.
C
C  Input:      x      := N x 1 vector
C         bounds      := N x 2 matrix, whose first column holds the
C                                      lower and whose second column
C                                       holds the upper bounds on x.
C
C  Returns:   .False.  if no component of x exceeds its respective
C              and .True. otherwise.
C

      Logical Function CheckBounds1(N,x,bounds)

      Integer(4) N, count
	Real(8) x(N), bounds(1:N,1:2)
	
      if (count(x .lt. bounds(1:N,1)) .gt. 0) then
	   CheckBounds1=.True.
	else
	   if (count(x .gt. bounds(1:N,2)) .gt. 0) then
	     CheckBounds1=.True.
	   else
	     CheckBound1=.False.
	   endif
	endif

      End Function
C********************************************************************
C CheckBounds2
C
C Purpose: Reduce the Newton-Raphson step so that
C          the point x_0 + s*dx remains in the domain
C          of the vector valued function f
C
C Usage:   s=CheckBound2(n,x,dx,bounds)
C
C Input:    n := integer(4), the number of variables
C           x := n times 1 real(8) vector, the inital point
C          dx := n times 1 real(8) vector, the Newton-Raphson step
C      bounds := n times 2 real(8) matrix , whose first (second) column
C                holds the lower (upper) bounds for x
C Output    s := Real(8) scalar
C

      Real(8) Function CheckBounds2(n,x,dx,bounds)

      Integer(4) n, i
	Real(8) x(1:n), dx(1:n), bounds(1:n,1:2), l1(1:n), l2(1:n), s1, s2

	Do i=1,n
	   l1(i)=(bounds(i,1)-x(i))/dx(i)
	   l2(i)=(bounds(i,2)-x(i))/dx(i)
	End Do
      s1=MinVal(l1,mask=l1 .gt. 0.0)
	s2=MinVal(l2,mask=l2 .gt. 0.0)

      s1=Min(s1,s2)
	if (s1 .lt. 1.0) then
        CheckBounds2=0.95*s1
	else
	  CheckBounds2=1.0
	endif

	Return

	End Function	


C********************************************************************
C
C Corrx
C
C Purpose: Compute correlation matrix vom data maxtrix
C
C Usage:   Call Corrx(n,m,x,crx)
C
C Input:   n, Integer(4), number of observations in x
C          m, Integer(4), number of variables (columns) in x
C          x, Real(8), n x m data matrix
C
C Output:  crx: m x m correlation matrix with standard
C               deviations on the diagonal
C
      Subroutine Corrx(n,m,x,crx)

      use IMSLF90 
	Integer(4) n, m
	Real(8) x(n,m), crx(m,m)
	
	Real(8) xbar(n,m), means(m), moment(m,m), stdv(m)
	Integer(4) I, J

C  Compute moment matrix
      means=Sum(x,dim=1)/n
	Do I=1,N
	   Do J=1,M
	    xbar(i,j)=x(i,j)-means(j)
	   End Do
	End Do
      moment=MatMul(Transpose(xbar),xbar)/(n-1)
      stdv=dsqrt(Diagonals(moment))
	Do I=1,M	   
	   crx(I,I)=stdv(i)
	End Do

	Do I=1,M
	   Do J=I+1,M
	      crx(i,j)=moment(i,j)/(stdv(i)*stdv(j))
	   End Do
	End Do

	End Subroutine

C********************************************************************
C
C FixpMN
C
C Purpose: find the zero of f(x) within prespecified bounds
C
C Usage: CALL FixpMN(x0,bounds,F,x)
C
C Input:   x0, Real(8), starting value
C      bounds, Real(8), 2 x 1, vector, lower and upper bounds for x
C           F, pointer to the function that computes f(x)
C
C Ouptut:   x, Real(8), the solution
C
     

      Subroutine FixpMN(x0,bounds,F,x)

      use IMSLF90 
      
	Real(8) x0, bounds(1:2), x

	Integer MaxIt, itn, korder
	Real(8) stopc, crit, bgstep, tol, df, fx, dx,step,x1,eps,
     +        xscale
	

	External F

C	Initialize 
      maxit=5000     ! stop after 5000 Iterations 
      stopc=1.0e-8   ! stopping criterium 
	eps=0.0
	xscale=1.0
	korder=1
      bgstep=0.01*abs(x0)
	tol=1.e-10
	
	x1=x0
      itn=1           
C Start Iterations
      crit=1.0
      do while ((crit .gt. stopc) .and. (itn .lt. maxit))
C	   df=DDERIV(F,korder,x1,bgstep,tol)
         Call F(1,x1,fx)
         Call DFDGRD(F,korder,x1,xscale,fx,eps,df)
C         dx=-F(x1)/df
         dx=-fx/df
	   x=x1+dx
         step=1
         do while ((x .lt. Bounds(1)) .or. (x .gt. bounds(2)))
          step=0.5*step         
          x=x1+step*dx
         End Do   
	   Call F(1,x,fx)      
	   crit=abs(fx)	   
         x1=x
         itn=itn+1
      End Do
      if (itn .gt. maxit) then
	   print *, 'Maximum number of iterations exceeded'
	   print *, 'Returned solution may be missleading'
      endif

      End Subroutine

C**********************************************************************
C
C MarkovAR
C
C Purpose: Approximate AR-Process by Markov Chain
C
C Usage: Call MarkovAR(size,m,rho,sigma,z,p)
C
C Input: size, scalar, Real(8), fraction of
C                      unconditional variance of AR process,
C                      determines grid size
C
C        m, Integer(4), number of states of Markov Chain
C
C      rho, Real(8), autoregressiv parameter
C  
C    sigma, Real(8), standard deviation of innovations
C
C Output: z, m x 1, Real(8) vector, states of the Markov Chain
C
C         p, m x m, Real(8) transition matrix
C
      Subroutine MarkovAR(size,m,rho,sigma,z,p)

      use IMSLF90       

	Integer(4) m
	Real(8) size, rho, sigma, z(1:m), p(1:m,1:m)

      Real(8) sigmaz, zbar
	Integer(4) I
		 
      sigmaz=dsqrt(sigma**2/(1-(rho**2)))
      zbar=size*sigmaz

      Call SeqaD(-zbar,dble(2.0*zbar/(m-1)), m,z)
  
      Do I=1,m
	
         p(i,1)=dnordf((z(1)-rho*z(i)+(z(2)-z(1))/2)/sigma)
C         p(i,m)=1-dnordf((z(m)-rho*z(i)-(z(m)-z(m-1))/2)/sigma)
         
	   Do j=2,m-1
   
           p(i,j)=dnordf((z(j)-rho*z(i)+(z(j)-z(j-1))/2)/sigma)-
     +             dnordf((z(j)-rho*z(i)-(z(j)-z(j-1))/2)/sigma)
         
	   End Do
	   p(i,m)=1-SUM(p(i,1:m-1))
	   
	End Do  
   
      End Subroutine

C*********************************************************************
C
C MarkovTP
C
C Purpose: Compute realization of Markov chain
C
C Usage: Call MarkovTP(n,m,p,i,zt)
C
C Input: n, Integer(4), length of time path
C    
C        m, Integer(4), size of transition matrix
C
C        p, Real(8), m x m Transition matrix
C
C        i0, Integer(4), index of first realization
C
C Ouptut: zt, Integer(4), n x 1 vector of indices from
C             the set 1, 2, 3, ..., m
C

      Subroutine MarkovTP(n,m,p,i0,zt)

      use IMSLF90       

	Integer(4) n, m, zt(1:n)
	Real(8) p(1:m,1:m)

      Integer i, j, l
      Real(8) eps(1:n), pu, po


C Draw a sequence of uniformly distributed random numbers
      Call DRNUN(n,eps)
 
      zt(1)=i0
	
	Do 10, I=1,N-1	
       j=zt(i)	 
       if (eps(i) .le. p(j,1)) then
             zt(i+1)=1	
             goto 10
       endif		   	       
       pu=p(j,1)
	 po=pu
	 Do 20, l=2,m
                 pu=po
                 po=pu+p(j,l)	
                 if ((eps(i) .gt. pu) .and. (eps(i) .le.po)) then
		         zt(i+1)=l
		         goto 20
			   endif
20    continue      

10    Continue

      End Subroutine